from spotdl.search.spotifyClient import get_spotify_client
from spotdl.search.songObj import SongObj

from typing import List

def search_for_song(query: str) ->  SongObj:
    '''
    `str` `query` : what you'd type into spotify's search box

    Queries Spotify for a song and returns the best match
    '''

    # get a spotify client
    spotifyClient = get_spotify_client()

    # get possible matches from spotify
    result = spotifyClient.search(query, type = 'track')

    # return first result link or if no matches are found, raise and Exception
    if len(result['tracks']['items']) == 0:
        raise Exception('No song matches found on Spotify')
    else:
        for songResult in result['tracks']['items']:
            songUrl = 'http://open.spotify.com/track/' + songResult['id']
            song = SongObj.from_url(songUrl)
            
            if song.get_youtube_link() != None:
                return song
        
        raise Exception('Could not match any of the results on YouTube')
        
def get_album_tracks(albumUrl: str) -> List[SongObj]:
    '''
    `str` `albumUrl` : Spotify Url of the album whose tracks are to be
    retrieved

    returns a `list<songObj>` containing Url's of each track in the given album
    '''

    spotifyClient = get_spotify_client()
    albumTracks = []

    trackResponse = spotifyClient.album_tracks(albumUrl)

    # while loop acts like do-while
    while True:

        for track in trackResponse['items']:
            song = SongObj.from_url('https://open.spotify.com/track/' + track['id'])
            
            if song.get_youtube_link() != None:
                albumTracks.append(song)
        
        # check if more tracks are to be passed
        if trackResponse['next']:
            trackResponse = spotifyClient.album_tracks(
                albumUrl,
                offset = len(albumTracks)
            )
        else:
            break
    
    return albumTracks

def get_playlist_tracks(playlistUrl: str) -> List[SongObj]:
    '''
    `str` `playlistUrl` : Spotify Url of the album whose tracks are to be
    retrieved

    returns a `list<songObj>` containing Url's of each track in the given playlist
    '''

    spotifyClient = get_spotify_client()
    playlistTracks = []

    playlistResponse = spotifyClient.playlist_tracks(playlistUrl)

    # while loop to mimic do-while
    while True:

        for songEntry in playlistResponse['items']:
            song = SongObj.from_url(
                'https://open.spotify.com/track/' + songEntry['track']['id'])
            
            if song.get_youtube_link() != None:
                playlistTracks.append(song)

        # check if more tracks are to be passed        
        if playlistResponse['next']:
            playlistResponse = spotifyClient.playlist_tracks(
                playlistUrl,
                offset = len(playlistTracks)
            )
        else:
            break
    
    return playlistTracks


#===============
#=== Imports ===
#===============

from spotipy import Spotify
from spotipy.oauth2 import SpotifyClientCredentials



#===============
#=== Globals ===
#===============

#! Look through both initialize() and get_spotify_client() for need of masterClient
masterClient = None



#=========================
#=== The actual action ===
#=========================

def initialize(clientId: str, clientSecret: str):
    '''
    `str` `clientId` : client id from your spotify account
    
    `str` `clientSecret` : client secret for your client id

    RETURNS `~`

    creates and caches a spotify client iff. a client doesn't exist. Can only be called
    once, multiple calls will cause an Exception.
    '''

    # check if initialization has been completed, if yes, raise an Exception
    #! None evaluates to False, objects evaluate to True.
    global masterClient

    if masterClient:
        raise Exception('A spotify client has already been initialized')



    # else create and cache a spotify client
    else:
        # create Oauth credentials for the SpotifyClient
        credentialManager = SpotifyClientCredentials(
            client_id = clientId,
            client_secret = clientSecret
        )

        client = Spotify(client_credentials_manager = credentialManager)
        
        masterClient = client

def get_spotify_client():
    '''
    RETURNS `Spotify`

    returns a cached spotify client of type `spotipy.Spotify`, can only be called after a
    call to `spotifyClient.initialize(clientId, clientSecret)`
    '''

    global masterClient
    
    #! None evaluvates to False, Objects evaluate to True
    if masterClient:
        return masterClient

    else:
        raise Exception('Spotify client not created. Call spotifyClient.initialize' +
            '(clientId, clientSecret) first.')


from setuptools import setup

long_desc = open('README.md', encoding = 'utf-8').read()

setup(
    # 'spotify-downloader' was already taken (＞﹏＜)
    name = "spotdl",

    packages = [
        'spotdl',
        'spotdl.search',
        'spotdl.download',
    ],

    #! PyPi follows Semantic Versioning - http://semver.org/
    version = '3.2.1',

    install_requires = [
        'spotipy',
        'pytube',
        'tqdm',
        'rapidfuzz',
        'requests',
        'mutagen',
        'ytmusicapi',
    ],

    extras_require={
        "test": [
            "pytest>=6.0",
            "pytest-mock==3.3.1",
            "pytest-vcr==1.0.2",
            "pyfakefs==4.3.0",
            "pytest-cov==2.10.1"
        ],
        "dev": [
            "tox"
        ]
    },

    description="Downloads Spotify music from Youtube with metadata and album art",
    long_description=long_desc,
    long_description_content_type='text/markdown',

    author="spotDL Team",
    author_email="unrealengin71+PyPi@gmail.com",

    license="MIT",

    python_requires=">=3.6",

    url="https://github.com/spotDL/spotify-downloader",
    download_url="https://pypi.org/project/spotdl/",

    keywords=[
        "spotify",
        "downloader",
        "download",
        "music",
        "youtube",
        "mp3",
        "album",
        "metadata",
    ],

    classifiers=[
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Utilities",
    ],

    entry_points = {
        "console_scripts": ["spotdl = spotdl.__main__:console_entry_point"]
    }
)
